'''
  Module for usefull decorators e.g. for logging function calls, input and output.
'''

import sys
import inspect
import logging
import threading
from time import time
from io import StringIO

#
# Help functions adopted from Michele Simionato's decorator module 
# http://www.phyast.pitt.edu/~micheles/python/decorator.zip
#

def getinfo(func):
    """
      Returns an info dictionary containing:
      - name (the name of the function : str)
      - argnames (the names of the arguments : list)
      - defaults (the values of the default arguments : tuple)
      - signature (the signature : str)
      - doc (the docstring : str)
      - module (the module name : str)
      - dict (the function __dict__ : str)
    """
    assert inspect.ismethod(func) or inspect.isfunction(func)
    signature=inspect.signature(func)
    output=dict(name=func.__name__, argnames=signature.parameters.keys(), signature=str(signature),
                defaults=func.__defaults__, doc=func.__doc__,
                module=func.__module__, dict=func.__dict__,
                globals=func.__globals__, closure=func.__closure__)
    return output

def update_wrapper(wrapper, wrapped, create=False):
    """
      An improvement over functools.update_wrapper. By default it works the
      same, but if the 'create' flag is set, generates a copy of the wrapper
      with the right signature and update the copy, not the original.
      Moreovoer, 'wrapped' can be a dictionary with keys 'name', 'doc', 'module',
      'dict', 'defaults'.
    """
    if isinstance(wrapped, dict):
        infodict=wrapped
    else:  # assume wrapped is a function
        infodict=getinfo(wrapped)
    assert not '_wrapper_' in infodict["argnames"], \
        '"_wrapper_" is a reserved argument name!'
    if create:  # create a brand new wrapper with the right signature
        src="lambda %(signature)s: _wrapper_(%(signature)s)"%infodict
        # import sys; print >> sys.stderr, src # for debugging purposes
        wrapper=eval(src, dict(_wrapper_=wrapper))
    try:
        wrapper.__name__=infodict['name']
    except:  # Python version < 2.4
        pass
    wrapper.__doc__=infodict['doc']
    wrapper.__module__=infodict['module']
    wrapper.__dict__.update(infodict['dict'])
    wrapper.__defaults__=infodict['defaults']
    return wrapper

# the real meat is here
def _decorator(caller, func):
    infodict=getinfo(func)
    argnames=infodict['argnames']
    assert not ('_call_' in argnames or '_func_' in argnames), \
        'You cannot use _call_ or _func_ as argument names!'
    src="lambda %(signature)s: _call_(_func_, %(signature)s)"%infodict
    dec_func=eval(src, dict(_func_=func, _call_=caller))
    return update_wrapper(dec_func, func)

def decorator(caller):
    """
    General purpose decorator factory: takes a caller function as
    input and returns a decorator with the same attributes.
    """
    return update_wrapper(lambda f: _decorator(caller, f), caller)

########################### Decorators for logging ##################################

def _nicetype(item):
    if 'array' in type(item).__name__:
        return type(item).__name__+("[%s]"%str(item.shape))
    elif type(item) in [list, tuple]:
        return type(item).__name__+('[%i]'%len(item))
    else:
        return type(item).__name__

class DecoratorLogger(logging.getLoggerClass()):
    '''
      A logger that makes sure the actual function definition filename, lineno and function name
      is used for logging.
    '''

    if sys.version_info[0:2]>=(3, 2):  # sinfo was introduced in python 3.2
        def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
            if extra is None:
                return logging.getLoggerClass().makeRecord(self, name, lvl, fn, lno,
                                                           msg, args, exc_info, func=func, extra=None, sinfo=sinfo)
            else:
                return logging.getLoggerClass().makeRecord(self, name, lvl, extra['name'], extra['lno'],
                                                           msg, args, exc_info, func=extra['func'], extra=None,
                                                           sinfo=sinfo)
    else:
        def makeRecord(self, name, lvl, fn, lno, msg, args, exc_info, func=None, extra=None):
            if extra is None:
                return logging.getLoggerClass().makeRecord(self, name, lvl, fn, lno,
                                                           msg, args, exc_info, func=func, extra=None)
            else:
                return logging.getLoggerClass().makeRecord(self, name, lvl, extra['name'], extra['lno'],
                                                           msg, args, exc_info, func=extra['func'], extra=None)

# only use the debug decorators in debugging runs.
old_class=logging.getLoggerClass()
logging.setLoggerClass(DecoratorLogger)
logger=logging.getLogger('deco')
logger.setLevel(logging.DEBUG)
null_handler=logging.StreamHandler(StringIO())
null_handler.setLevel(logging.CRITICAL)
logger.addHandler(null_handler)
logging.setLoggerClass(old_class)

def _logformat(msg, decname, func):
    if hasattr(func, 'im_func'):
        co=func.__func__.__code__
    else:
        co=func.__code__
    fname=co.co_filename
    lno=co.co_firstlineno
    logger.debug(msg, extra={'func': '@'+decname, 'name': fname, 'lno': lno})

@decorator
def log_call(func, *args, **kw):
    '''
      Decorator to log just the method call.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        _logformat('%s.%s.%s <%s>'%(infodict['module'],
                                    args[0].__class__.__name__, infodict['name'], repr(threading.current_thread())),
                   'log_call', func)
    else:
        _logformat('%s.%s <%s>'%(infodict['module'], infodict['name'], repr(threading.current_thread())), 'log_call',
                   func)
    return func(*args, **kw)

@decorator
def log_input(func, *args, **kw):
    '''
      Decorator to log a method call with input.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if hasattr(func, 'im_func'):
        logstr=' call %s.%s('%(args[0].__class__.__name__, infodict['name'])
        method=True
        info_len=len('%s.%s('%(args[0].__class__.__name__, infodict['name']))
    else:
        logstr='call %s('%(infodict['name'])
        method=False
        info_len=len('%s('%(infodict['name']))

    namelen=max([len(name)+3 for name in infodict['argnames']]+[len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i==0 and (method or infodict['name']=='__init__'):
            continue
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+7))
        logstr+='\n'+('%% %is= %%s (%%s)'%(namelen+3))%(infodict['argnames'][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+4))
        logstr+='\n'+('KW %% %ss= %%s (%%s)'%namelen)%(key, repr(value), _nicetype(arg))
    logstr+='\n)'
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    logstr=logstr.replace('call ', 'call \n'+' '*10)

    _logformat(logstr, 'log_input', func)
    return func(*args, **kw)

@decorator
def log_output(func, *args, **kw):
    '''
      Decorator to log a method call with output. If combined with log_input
      the input is logged at the time before the call and the output after.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    output=func(*args, **kw)
    infodict=getinfo(func)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        logstr='return from %s.%s'%(args[0].__class__.__name__, infodict['name'])
    else:
        logstr='return from %s'%(infodict['name'])
    value=repr(output)
    value_split=value.splitlines()
    if len(value_split)>5:
        value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
    logstr+='\n-> %15s (%s)'%(value, _nicetype(output))
    logstr=logstr.replace('\n', '\n'+' '*44)
    _logformat(logstr, 'log_output', func)
    return output

@decorator
def log_both(func, *args, **kw):
    '''
      Decoratore to log a method call with input and output.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    infodict=getinfo(func)
    if hasattr(func, 'im_func'):
        logstr=' call %s.%s('%(args[0].__class__.__name__, infodict['name'])
        method=True
        info_len=len('%s.%s('%(args[0].__class__.__name__, infodict['name']))
    else:
        logstr='call %s('%(infodict['name'])
        method=False
        info_len=len('%s('%(infodict['name']))

    namelen=max([len(name)+3 for name in infodict['argnames']]+[len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i==0 and (method or infodict['name']=='__init__'):
            continue
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+7))
        logstr+='\n'+('%% %is= %%s (%%s)'%(namelen+3))%(infodict['argnames'][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value=repr(arg)
        value_split=value.splitlines()
        if len(value_split)>5:
            value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
        value.replace('\n', '\n'+' '*int(namelen+4))
        logstr+='\n'+('KW %% %ss= %%s (%%s)'%namelen)%(key, repr(value), _nicetype(arg))
    logstr+='\n)'
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    logstr=logstr.replace('call ', 'call \n'+' '*10)

    _logformat(logstr, 'log_input', func)
    # call the function
    output=func(*args, **kw)
    if len(infodict['argnames'])>0 and infodict['argnames'][0]=='self':
        logstr='return from %s.%s'%(args[0].__class__.__name__, infodict['name'])
    else:
        logstr='return from %s'%(infodict['name'])
    value=repr(output)
    value_split=value.splitlines()
    if len(value_split)>5:
        value="\n".join(value_split[:2]+[' '*int(len(value_split[1])/2)+'...']+value_split[-2:])
    logstr+='\n-> %s (%s)'%(value, _nicetype(output))
    logstr=logstr.replace('\n', '\n'+' '*int(info_len+10))
    _logformat(logstr, 'log_output', func)
    return output

timings={}

@decorator
def time_call(func, *args, **kw):
    '''
      Decorator to log just the method call.
    '''
    if logging.root.getEffectiveLevel()>logging.DEBUG: return func(*args, **kw)
    name=func.__name__
    start=time()
    output=func(*args, **kw)
    runtime=time()-start
    if not name in timings:
        timings[name]=(0., 0.)
    avg, calls=timings[name]
    timings[name]=((avg*calls+runtime)/(calls+1), calls+1)
    return output

########################## General decorators ###############################

class check_input(object):
    '''
      Decorator checking the input to a function.

      Usage:
      @check_input([type1, type2, ...], try_convert=True)
      def your_function(var1, var2, ...):
          ... function text ...
    '''
    try_convert=True
    types=[]

    def __init__(self, types, try_convert=True):
        self.types=types
        self.try_convert=try_convert

    def __call__(self, func):
        infodict=getinfo(func)
        argnames=infodict['argnames']
        assert not ('_call_' in argnames or '_func_' in argnames), \
            'You cannot use _call_ or _func_ as argument names!'
        src="def dec_func(%(signature)s):"%infodict
        for i, typei in enumerate(self.types):
            src+='\n  if type(%s).__name__!="%s":'%(argnames[i], typei.__name__)
            if self.try_convert:
                src+='\n    try:'
                src+='\n      %s=%s(%s)'%(argnames[i], typei.__name__, argnames[i])
                src+='\n    except:'
                src+='\n      raise ValueError("type of %s is not %s")'%(argnames[i], typei.__name__)
            else:
                src+='\n    raise ValueError("type of %s is not %s")'%(argnames[i], typei.__name__)
        src+='\n  return _func_(%(signature)s)'%infodict
        exec_dict=dict(_func_=func, _call_=self.__call__)
        exec(src, exec_dict)
        return update_wrapper(exec_dict['dec_func'], func)



import math
import numba
from numpy import empty, ndarray, empty_like

pi=math.pi
rad=pi/180.
sqrt2=math.sqrt(2.)

@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64, numba.float64, numba.float64),
    nopython=True, parallel=True, cache=True)
def GaussIntensity(alpha, s1, s2, sigma_x):
    I=empty_like(alpha)
    if s1==s2:
        for ai in numba.prange(alpha.shape[0]):
            sinalpha=math.sin(alpha[ai]*rad)
            I[ai]=math.erf(s2/sqrt2/sigma_x*sinalpha)
    else:
        for ai in numba.prange(alpha.shape[0]):
            sinalpha=math.sin(alpha[ai]*rad)
            I[ai]=(math.erf(s1/sqrt2/sigma_x*sinalpha)+math.erf(s2/sqrt2/sigma_x*sinalpha))/2.
    return I

@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64, numba.float64),
    nopython=True, parallel=True, cache=True)
def SquareIntensity(alpha, slen, beamwidth):
    I=empty_like(alpha)
    scale=slen/beamwidth
    for ai in numba.prange(alpha.shape[0]):
        F=scale*math.sin(alpha[ai]*rad)
        I[ai]=min(1.0, F)
    return I

@numba.jit(
    numba.float64[:](numba.float64, numba.float64[:]),
    nopython=True, parallel=True, cache=True)
def TwoThetatoQScalar(wavelength, TwoTheta):
    Q=empty_like(TwoTheta)
    scale=4.*pi/wavelength
    thscale=rad/2.0
    for li in numba.prange(Q.shape[0]):
        Q[li]=math.sin(TwoTheta[li]*thscale)*scale
    return Q

@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64[:]),
    nopython=True, parallel=True, cache=True)
def TwoThetatoQVector(wavelength, TwoTheta):
    Q=empty_like(TwoTheta)
    scale=4.*pi
    thscale=rad/2.0
    for li in numba.prange(Q.shape[0]):
        Q[li]=math.sin(TwoTheta[li]*thscale)*scale/wavelength[li]
    return Q

def TwoThetatoQ(wavelength, TwoTheta):
    if type(wavelength) is ndarray:
        return TwoThetatoQVector(wavelength, TwoTheta)
    else:
        return TwoThetatoQScalar(wavelength, TwoTheta)

@numba.jit(
    numba.float64[:](numba.float64, numba.float64[:]),
    nopython=True, parallel=True, cache=True)
def QtoThetaScalar(wavelength, Q):
    Theta=empty_like(Q)
    scale=wavelength/4.0/pi
    for li in numba.prange(Q.shape[0]):
        Theta[li]=math.asin(scale*Q[li])/rad
    return Theta

@numba.jit(
    numba.float64[:](numba.float64[:], numba.float64[:]),
    nopython=True, parallel=True, cache=True)
def QtoThetaVector(wavelength, Q):
    Theta=empty_like(Q)
    scale=1.0/4.0/pi
    for li in numba.prange(Q.shape[0]):
        Theta[li]=math.asin(scale*wavelength[li]*Q[li])/rad
    return Theta

def QtoTheta(wavelength, Q):
    if type(wavelength) is ndarray:
        return QtoThetaVector(wavelength, Q)
    else:
        return QtoThetaScalar(wavelength, Q)

#
@numba.jit(
    numba.types.Tuple((numba.float64[:], numba.float64[:, :]))(numba.float64[:], numba.float64, numba.int16,
                                                               numba.float64),
    nopython=True, parallel=True, cache=True)
def ResolutionVectorScalar(Q, dQ, points, range):
    Qres=empty(Q.shape[0]*points)
    weight=empty((points, Q.shape[0]))
    Qstep=2.0*range/points*dQ
    NQ=Q.shape[0]

    wscale=1.0/math.sqrt(2.0*pi)/dQ

    for ri in numba.prange(points):
        dq=(ri-(points-1.0)/2.0)*Qstep
        for qi in numba.prange(NQ):
            Qi=Q[qi]
            Qj=Qi+dq
            Qres[qi+ri*NQ]=Qj
            weight[ri, qi]=wscale*math.exp(-(Qi-Qj)**2/(dQ)**2/2.0)

    return (Qres, weight)

@numba.jit(
    numba.types.Tuple((numba.float64[:], numba.float64[:, :]))(numba.float64[:], numba.float64[:], numba.int16,
                                                               numba.float64),
    nopython=True, parallel=True, cache=True)
def ResolutionVectorVector(Q, dQ, points, range):
    Qres=empty(Q.shape[0]*points)
    weight=empty((points, Q.shape[0]))
    Qstep_scale=2.0*range/points
    NQ=Q.shape[0]

    wscale=1.0/math.sqrt(2.0*pi)

    for ri in numba.prange(points):
        dq_scale=(ri-(points-1.0)/2.0)*Qstep_scale
        for qi in numba.prange(NQ):
            dq=dq_scale*dQ[qi]
            Qi=Q[qi]
            Qj=Qi+dq
            Qres[qi+ri*NQ]=Qj
            weight[ri, qi]=wscale*math.exp(-(Qi-Qj)**2/(dQ[qi])**2/2.0)/dQ[qi]

    return (Qres, weight)

def ResolutionVector(Q, dQ, points, range=3):
    if type(dQ) is ndarray:
        return ResolutionVectorVector(Q, dQ, points, range)
    else:
        return ResolutionVectorScalar(Q, dQ, points, range)



'''<h1> Figure of Merit (FOM)</h1>
The Figure of Merit (FOM) is the function that compares how well the simulation matches the measured data. Strictly speaking, for Gaussian errors, a chi squared (&chi;<sup>2</sup>) FOM is the most appropriate. However, the world is not perfect and many times the data can be fitted more easily and more robustly if another FOM is chosen. Each FOM function has its merits and drawbacks, and fitting can rely critically on choosing the right FOM function for the particular data to be analyzed. The following gives a brief summary and explanation of the FOMs included in the standard GenX distribution so far.<br>
It is also possible to create custom FOM functions to be used by GenX. For more information on this refer to the Section "Customization" below.<br>


<h2>Available FOM functions</h2>
In the following, the merged data set consisting of all data sets
that are marked for use is denoted as <var>Y</var> and the corresponding
simulation is denoted as <var>S</var>. A single element of these arrays
is indicated by a subscript <var>i</var>. In the same manner, the
independent variable (denoted as <var>x</var> in the data strucure) is called
<var>X</var>. The error array is denoted <var>E</var>. Finally the total number
of data points is given by <var>N</var> and <var>p</p> is the number of free parameters
in the fit.<br>


<h3>Unweighted FOM functions</h3>


<h4>diff</h4>
Average of the absolute difference between simulation and data.<br>
<br><huge>
    FOM<sub>diff</sub> =  1/(N-p) &times; &#8721;<sub><var>i</var></sub>
    &#124;<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>&#124;
</huge><br>


<h4>log</h4>
Average of the absolute difference between the logarithms (base 10) of the data and the simulation.<br>
<br><huge>
    FOM<sub>log</sub> = 1/(N-p) &times;&#8721;<sub><var>i</var></sub>
    &#124;log<sub>10</sub>(<var>Y<sub>i</sub></var>) -
    log<sub>10</sub>(<var>S<sub>i</sub></var>)&#124;
</huge><br>


<h4>sqrt</h4>
Average of the absolute difference between the square roots of the data and the simulation:<br>
<br><huge>
    FOM<sub>sqrt</sub> =  1/(N-p) &times; &#8721;<sub><var>i</var></sub>
    &#124;sqrt(<var>Y<sub>i</sub></var>) - sqrt(<var>S<sub>i</sub></var>)
    &#124;
</huge><br>


<h4>R1</h4>
Crystallographic R-factor (often denoted as R1, sometimes called residual factor or reliability factor or the R-value or R<sub>work</sub>).<br>
Gives the percentage of the summed structure factor residuals (absolute difference between data and simulation) over the entire data set with respect to the total sum of measured structure factors. For data sets spanning several orders of magnitude in intensity, R1 is dominated by the residuals at high intensities, while large residuals at low intensities have very little impact on R1.
This implementation here assumes that the loaded data are intensities (squares of the structure factors), hence the square roots of the loaded data are taken for the calculation of R1.<br>
[A.J.C. Wilson, Acta Crystallogr. A32, 994 (1976)]<br>
<br><huge>
  FOM<sub>R1</sub> =
  &#8721;<sub><var>i</var></sub> [ 
  &#124;sqrt(<var>Y<sub>i</sub></var>) - sqrt(<var>S<sub>i</sub></var>)
  &#124; ] / &#8721;<sub><var>i</var></sub> [ sqrt(<var>Y<sub>i</sub></var>) ]
</huge><br>


<h4>logR1</h4>
The logarithmic R1 factor is a modification of the crystallographic R-factor, calculated using the logarithm (base 10) of the structure factor and simulation. This scaling results in a more equal weighting of high-intensity and low-intensity data points which can be very helpful when fitting data which is spanning several orders of magnitude on the y-axis. Essentially it gives all data points equal weight when displayed in a log-plot.<br>
<br><huge>
    FOM<sub>logR1</sub> =
    &#8721;<sub><var>i</var></sub> [ &#124;
    log<sub>10</sub>(sqrt(<var>Y<sub>i</sub></var>)) -
    log<sub>10</sub>(sqrt(<var>S<sub>i</sub></var>))
    &#124; ] / 
    &#8721;<sub><var>i</var></sub> [
    log<sub>10</sub>(sqrt(<var>Y<sub>i</sub></var>) ]
</huge><br>


<h4>R2</h4>
Crystallographic R2 factor. In contrast to R1, this gives the ratio of the total sum of squared deviations to the total sum of squared structure factors. (Note that sometimes R2 is also defined as the square root of the value defined here.)
Like in the case for R1, this implementation assumes that the loaded data are intensities (squares of the structure factors).<br>
[A.J.C. Wilson, Acta Crystallogr. A32, 994 (1976)]<br>
<br><huge>
    FOM<sub>R2</sub> =
    &#8721;<sub><var>i</var></sub> [
    (<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>)<sup>2</sup> ] /
    &#8721;<sub><var>i</var></sub> [ <var>Y<sub>i</sub><sup>2</sup></var> ]
</huge><br>


<h4>logR2</h4>
The logarithmic R2 factor is a modification of the crystallographic R2 factor, calculated using the logarithm (base 10) of the structure factor and simulation. This scaling results in a more similar weighting of high-intensity and low-intensity data points which can be very helpful when fitting data which is spanning several orders of magnitude on the y-axis. Essentially it gives all data points equal weight when displayed in a log-plot.<br>
<br><huge>
    FOM<sub>logR2</sub> =
    &#8721;<sub><var>i</var></sub> [ 
    (log<sub>10</sub>(<var>Y<sub>i</sub></var>) -
    log<sub>10</sub>(<var>S<sub>i</sub></var>)
    )<sup>2</sup> ] /
    &#8721;<sub><var>i</var></sub> [ 
    log<sub>10</sub>(<var>Y<sub>i</sub>)<sup>2</sup></var> ]
</huge><br>


<h4>sintth4</h4>
Gives the average of the absolute differences scaled with a sin(2&theta;)<sup>4</sup> term (2&theta; = tth). For reflectivity data, this will divide away the Fresnel reflectivity. <br>
<br><huge>
    FOM<sub>sintth4</sub> = 1/(N-p) &times;
    &#8721;<sub><var>i</var></sub>
    &#124;<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>&#124; &times;
    sin(<var>tth</var>)<sup>4</sup>
</huge><br>


<h3>Weighted FOM functions</h3>

<h4>chi2bars</h4>
Chi squared (&chi;<sup>2</sup>) FOM including error bars<br>
<br><huge>
    FOM<sub>chi2bars</sub> = 1/(N-p) &times; &#8721;<sub><var>i</var></sub>
    ((<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>) /
    <var>E<sub>i</sub></var>)<sup>2</sup>
</huge><br>


<h4>chibars</h4>
Chi squared but without the squaring! Includes error bars:<br>
<br><huge>
    FOM<sub>chibars</sub> = 1/(N-p) &times; &#8721;<sub><var>i</var></sub>
    &#124;(<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>) /
    <var>E<sub>i</sub></var>&#124;
</huge><br>


<h4>logbars</h4>
Absolute logarithmic (base 10) difference, taking errors into account:<br>
<br><huge>
    FOM<sub>logbars</sub> = 1/(N-p) &times; &#8721;<sub><var>i</var></sub>
    &#124;log<sub>10</sub>(<var>Y<sub>i</sub></var>) -
    log<sub>10</sub>(<var>S<sub>i</sub></var>)&#124; /
    <var>E<sub>i</sub></var>*ln(10)*<var>Y<sub>i</sub></var>
</huge><br>


<h4>R1bars</h4>
Similar to the crystallographic R-factor R1, but with weighting of the data points by the experimental error values. The error values in E are assumed to be proportional to the standard deviation of the measured intensities.<br>
[A.J.C. Wilson, Acta Crystallogr. A32, 994 (1976), W.C. Hamilton, Acta Crystallogr. 18(3), 502 (1965)]<br>
<br><huge>
    FOM<sub>R1bars</sub> =
    &#8721;<sub><var>i</var></sub><var> [ sqrt(1/E<sub>i</sub></var>) &times;
    &#124;sqrt(<var>Y<sub>i</sub></var>) - sqrt(<var>S<sub>i</sub></var>)
    &#124; ] /
    &#8721;<sub><var>i</var></sub> [ sqrt(1/E<sub>i</sub></var>) &times;
    sqrt(<var>Y<sub>i</sub></var>) ]
</huge><br>


<h4>R2bars</h4>
Weighted R2 factor. The error values in E are assumed to be proportional to the standard deviation of the measured intensities.<br>
[A.J.C. Wilson, Acta Crystallogr. A32, 994 (1976), W.C. Hamilton, Acta Crystallogr. 18(3), 502 (1965)]<br>
<br><huge>
    FOM<sub>R2bars</sub> =
    &#8721;<sub><var>i</var></sub> [ (1/E<sub>i</sub></var>) &times;
    (<var>Y<sub>i</sub></var> - <var>S<sub>i</sub></var>)<sup>2</sup> ] /
    &#8721;<sub><var>i</var></sub> [ (1/E<sub>i</sub></var>) &times;
    <var>Y<sub>i</sub><sup>2</sup></var> ]
</huge><br>


<h2>Customization</h2>
Users can add their own cumstom-built FOM functions to be used in GenX. For detailed instructions on how to write the code for a custom FOM function and how to include it in the list of FOM functions available to GenX, see the manual at
<a href = "http://apps.sourceforge.net/trac/genx/wiki/DocPages/WriteFom">
http://apps.sourceforge.net/trac/genx/wiki/DocPages/WriteFom </a>
'''
# ==============================================================================

import numpy as np

# import also the custom FOM functions defined in fom_funcs_custom.py
# (do nothing if file does not exist)
try:
    # noinspection PyUnresolvedReferences
    from fom_funcs_custom import *
    from .gui_logging import iprint
    iprint("Imported custom-defined FOM functions from fom_funcs_custom.py")
except:
    pass

# ==============================================================================
# BEGIN FOM function defintions

# =========================
# unweighted FOM functions

def diff(simulations, data):
    ''' Average absolute difference
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    # return 1.0/(N-1)*np.sum([np.sum(np.abs(dataset.y - sim))\
    #    for (dataset, sim) in zip(data,simulations) if dataset.use])
    return [(dataset.y-sim)
            for (dataset, sim) in zip(data, simulations)]

diff.__div_dof__=True

def log(simulations, data):
    ''' Average absolute logartihmic difference
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [(np.log10(dataset.y)-np.log10(sim))
            for (dataset, sim) in zip(data, simulations)]

log.__div_dof__=True

def sqrt(simulations, data):
    ''' Average absolute difference of the square root
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [(np.sqrt(dataset.y)-np.sqrt(sim))
            for (dataset, sim) in zip(data, simulations)]

sqrt.__div_dof__=True

def R1(simulations, data):
    ''' Crystallographic R-factor (R1)
    '''
    denom=np.sum([np.sum(np.sqrt(np.abs(dataset.y))) for dataset in data \
                  if dataset.use])
    return [1.0/denom*(np.sqrt(np.abs(dataset.y))-np.sqrt(np.abs(sim))) \
            for (dataset, sim) in zip(data, simulations)]

def logR1(simulations, data):
    ''' logarithmic crystallographic R-factor (R1)
    '''
    denom=np.sum([np.sum(np.log10(np.sqrt(dataset.y))) for dataset in data \
                  if dataset.use])
    return [1.0/denom*(np.log10(np.sqrt(dataset.y))-
                       np.log10(np.sqrt(sim))) \
            for (dataset, sim) in zip(data, simulations)]

def R2(simulations, data):
    ''' Crystallographic R2 factor
    '''
    denom=np.sum([np.sum(dataset.y**2) for dataset in data \
                  if dataset.use])
    return [1.0/denom*np.sign(dataset.y-sim)*(dataset.y-sim)**2 \
            for (dataset, sim) in zip(data, simulations)]

def logR2(simulations, data):
    ''' logarithmic crystallographic R2 factor
    '''
    denom=np.sum([np.sum(np.log10(dataset.y)**2) for dataset in data \
                  if dataset.use])
    return [1.0/denom*np.sign(np.log10(dataset.y)-np.log10(sim))*(np.log10(dataset.y)-np.log10(sim))**2 \
            for (dataset, sim) in zip(data, simulations)]

def sintth4(simulations, data):
    ''' Sin(tth)^4 scaling of the average absolute difference for reflectivity.
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [np.sin(dataset.x*np.pi/360.0)**4*
            (dataset.y-sim)
            for (dataset, sim) in zip(data, simulations)]

sintth4.__div_dof__=True

def Norm(simulations, data):
    '''  dataset normalized 1/3 scaling of the error
    '''
    return [1.0/np.sum(np.abs(dataset.y))*(np.sign(dataset.y)*np.abs(dataset.y)-np.sign(sim)*np.abs(sim)) \
            for (dataset, sim) in zip(data, simulations)]

Norm.__div_dof__=True

# =======================
# weighted FOM functions

def chi2bars(simulations, data):
    ''' Weighted chi squared
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [np.sign(dataset.y-sim)*(dataset.y-sim)**2/dataset.error**2
            for (dataset, sim) in zip(data, simulations)]

chi2bars.__div_dof__=True

def chibars(simulations, data):
    ''' Weighted chi squared but without the squaring
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [((dataset.y-sim)/dataset.error)
            for (dataset, sim) in zip(data, simulations)]

chibars.__div_dof__=True

def logbars(simulations, data):
    ''' Weighted average absolute difference of the logarithm of the data
    '''
    N=np.sum([len(dataset.y)*dataset.use for dataset in data])
    return [((np.log10(dataset.y)-np.log10(sim))
             /dataset.error*np.log(10)*dataset.y)
            for (dataset, sim) in zip(data, simulations)]

logbars.__div_dof__=True

def R1bars(simulations, data):
    ''' Weighted crystallographic R-factor (R1)
    '''
    denom=np.sum([np.sum(np.sqrt(1/dataset.error)*np.sqrt(dataset.y))
                  for dataset in data if dataset.use])
    return [1.0/denom*np.sqrt(1/dataset.error)*
            (np.sqrt(dataset.y)-np.sqrt(sim))
            for (dataset, sim) in zip(data, simulations)]

def R2bars(simulations, data):
    ''' Weighted crystallographic R2 factor
    '''
    denom=np.sum([(1/dataset.error)*np.sum(dataset.y**2)
                  for dataset in data if dataset.use])
    return [1.0/denom*(1/dataset.error)*np.sign(dataset.y-sim)*(dataset.y-sim)**2
            for (dataset, sim) in zip(data, simulations)]

# END FOM function definition
# ==============================================================================


# create introspection variables so that everything updates automatically
# Find all objects in this namespace
# (this includes the custom-defined FOM functions from fom_funcs_custom.py)
obj_list=dir()[:]

# find all functions
all_func_names=[s for s in obj_list if type(eval(s)).__name__=='function']
func_names=[s for s in all_func_names if all_func_names[0]!='_']

# End of file
# ==============================================================================
